!*********************************************************
! Program for fitting PSF or 2-d gaussian
! functions to image data
!
!**********************************************************
module ompsfintegration_mod

    use dal
    use cal
    use param  
    use sasdal
    use errorHandling
    use cubicshepard_mod
    use gridInterpolation_mod
    use splinefit_mod
    use integration_mod
    use ominterpolate_mod
    implicit none 


type(structShepard),    save :: psfShepardFit
type(GridInterpolant),    save :: psfGridFit

type(TwoDSpline), save :: psfSplineFit
type(BilinearStruct), save :: bilinearData



 
contains

!*******************************************
! Interpolates psf function
! NB dx(2) = distance from psf x centre
!    dx(3) = distance from psf y centre
!    dx(4) = pixel height
!*******************************************
real(double) Function PSFPixel(x)
    real(double) :: x(:)
 

    PSFPixel =  max(0.0_double, GridInterpolate(x(2), x(3), PSFGridFit))

!  PSFPixel = max(0.0, bilinearInterpolate(real(x(2)), real(x(3)), bilinearData))


end function PSFPixel
!***************************************************
!
!**************************************************
real(double) Function PSFIntegration(dx, dy)
    real(double), intent(in) :: dx, dy
    
    PSFIntegration = DoubleIntegration(dx - 0.5_double, dx + 0.5_double, &
                        dy - 0.5_double, dy + 0.5_double, PSFFunction1)

end function PSFIntegration

!*****************************************************
!
!****************************************************
real(double) Function PSFFunction1(dx)
    real(double), intent(in) :: dx
    xSave = dx
    PSFFunction1 = QGauss1(ylSave, yUSave, ShepardPSF)

end function PSFFunction1

!***************************************************
!
!**************************************************
real(double) Function ShepardIntegration(dx, dy)
    real(double), intent(in) :: dx, dy
    
    ShepardIntegration = DoubleIntegration(dx - 0.5_double, dx + 0.5_double, &
                        dy - 0.5_double, dy + 0.5_double, ShepardPSFFunction1)

end function ShepardIntegration

!*****************************************************
!
!****************************************************
real(double) Function ShepardPSFFunction1(dx)
    real(double), intent(in) :: dx
    xSave = dx
    ShepardPSFFunction1 = QGauss1(ylSave, yUSave, ShepardPSF)

end function ShepardPSFFunction1

!*******************


!***************************************************
!
!**************************************************
real(double) Function ShepardPSF(dy)
    real(double), intent(in) :: dy

    ShepardPSF = ShepardInterpolate(psfShepardFit, xSave, dy)   

end function ShepardPSF
!*****************************************************
!
!****************************************************
real(double) Function GridPSFFunction2(dx)
    real(double), intent(in) :: dx
    xSave = dx
    GridPSFFunction2 = QGauss1(ylSave, yUSave, GridPSF)

end function GridPSFFunction2

!***************************************************
!
!**************************************************
real(double) Function GridIntegration1(dx, dy)
    real(double), intent(in) :: dx, dy
    
    GridIntegration1 = DoubleIntegration(dx - 0.5_double, dx + 0.5_double, &
                        dy - 0.5_double, dy + 0.5_double, GridPSFFunction2)

end function GridIntegration1

!***************************************************
!
!**************************************************
real(double) Function GridIntegration2(dx, dy, width, height)
    real(double), intent(in) :: dx, dy, width, height
    
    GridIntegration2 = DoubleIntegration(dx - width, dx + width, &
                        dy - height, dy + height, GridPSFFunction2)

end function GridIntegration2






!*****************************************************
!
!****************************************************
real(double) Function GridPSFFunction1(dx)
    real(double), intent(in) :: dx
    xSave = dx
    GridPSFFunction1 = QGauss1(ylSave, yUSave, GridPSF)

end function GridPSFFunction1

!*******************


!***************************************************
!
!**************************************************
real(double) Function GridPSF(dy)
    real(double), intent(in) :: dy

    GridPSF = GridInterpolate(xSave, dy, psfGridFit)   

end function GridPSF


!***************************************************
!
!**************************************************
real(double) Function PSF(dy)
    real(double), intent(in) :: dy

    PSF = InterpolateSpline(xSave, dy, psfSplineFit)   

end function PSF



!*******************************************
! Integrates over psf pixel
! 
!******************************************
real(double) Function PSFPixel1(x)
    real(double) :: x(:)
 
!    PSFPixel1 =  x(3) *  ShepardIntegration(x(1), x(2))
 PSFPixel1 =  x(3) *  GridInterpolate(x(1), x(2), psfGridFit)

end function PSFPixel1


!*******************************************
! Integrates over psf pixel
! 
!******************************************
real(double) Function PSFPixel2(x)
    real(double) :: x(:)
 

    PSFPixel2 =  x(3) *  GridInterpolate(x(1), x(2), psfGridFit)


end function PSFPixel2






end module ompsfintegration_mod





























