!******************************************************************
! Module OmdetectSkybackground_mod
!******************************************************************
! 2013-10-15 vny:
! modifying the calculation of the background count error: instead of sigma1 
! calculating sigma1/sqrt(n) where n is the number of averaged points
! (Subroutine BackgroundCountsPerPixel)
!---
! 2013-11-12 vny: adding a separate parameter (sigmaPoisson) to the 
! subroutine BackgroundCountsPerPixel, because changing the old
! parameter (sigma) results in a changed count rate values
!*************************************************************************
module OmSkybackground_mod
    use cal
    use dal     
    use omallocation_mod
    use omnumerics_mod
    use omsourceparam_mod
    use omfunctions_mod
    use brent_mod
    use nonlinls_mod
    use ommode_mod
    use brent_mod
interface BackgroundMode
    module procedure BackgroundMode1,  &
                     ComputeBackgroundMode,  &
                     ComputeBackgroundMode1, & 
                     ComputeBackgroundMode2, & 
                     ComputeBackgroundMode3, &
                     ComputeBackgroundMode4
end interface

interface BackgroundCountsPerPixel
    module procedure BackgroundCountsPerPixel1,  &
                     BackgroundCountsPerPixel2
end interface



integer(int8), private, parameter :: numberBackgroundMethods = 6_int8
integer(int8), private :: backGroundMethod = 1   ! 6 Default SAS background determination method
real(single), private :: backgroundNsigma = 2.0


real(double), private :: meanValue = 0.0
    
                    

character(100), private, dimension(6) :: backgroundText =      & 
       (/ "SAS                                              ", &   
          "Mean                                             ", &          
          "Median                                           ", &
          "Mode                                             ", &          
          "Robust Mean                                      ", &
          "Robust Median                                    "/)    
                    





contains

!*****************************************
! setBackgroundNsigma
!*****************************************
integer(int8) function getNumberBackgroundMethods()
   
    
    getNumberBackgroundMethods = numberBackgroundMethods

end function getNumberBackgroundMethods




!**********************************************************************************
!
!*********************************************************************************
function getBackgroundText( i ) result(text)
    integer, intent(in) :: i
    character( len(trim(backgroundText(i)))) :: text
    text = trim(backgroundText(i) )

end function getBackgroundText

!*****************************************
! setBackgroundMethod
!*****************************************
subroutine setBackgroundMethod( method )
    integer(int8), intent(in) :: method
    
    backgroundMethod = method

end subroutine setBackgroundMethod


!*****************************************
! setBackgroundNsigma
!*****************************************
subroutine setBackgroundNsigma( value )
    real(single), intent(in) :: value
    
    backgroundNsigma = value

end subroutine setBackgroundNsigma



!*****************************************
! backgroundCountsPerPixel: 2013-11-12 renaming to backgroundCountsPerPixel1
!*****************************************
! 
   real(single) function BackgroundCountsPerPixel1(nData, data, &
                       nSigma, sigma)
    integer(int32), intent(in) :: nData
    real(single) :: data(:)

  
    real(single), intent(in)  :: nSigma
    real(single), intent(out) :: sigma

    select case (backgroundMethod)
    case(1)   ! The current SAS
       
        backgroundCountsPerPixel1 = backgroundmode1(ndata, data, &
             0.001, backgroundnsigma, sigma)
            

    case(2)   ! Simple average

       call ComputeAverage2(data, ndata, backgroundCountsPerPixel1, sigma)
       

    case(3)   ! Simple median

        backgroundCountsPerPixel1 = MedianValue(data, nData)
        sigma = max(1.0, sqrt(backgroundCountsPerPixel1))
        
        
    case(4)   ! The mode
   

        backgroundCountsPerPixel1 = ommode(ndata, data)
        sigma = max(1.0, sqrt(backgroundCountsPerPixel1))
        

    case(5)   ! Robust mean
       
       backgroundCountsPerPixel1 = RobustMean(data, nData, sigma)
              

    case(6) ! Robust median

        backgroundCountsPerPixel1 = RobustMedian(data, nData)
        sigma = max(1.0, sqrt(backgroundCountsPerPixel1))
        
        
    case default
        if(nsigma < 0.0) then
        end if
    end select


end function BackgroundCountsPerPixel1


!*****************************************
! backgroundCountsPerPixel2: 2013-11-12 renaming to backgroundCountsPerPixel1
!*****************************************
! 2013-11-12 vny: adding a new parameter sigmaPoisson
   real(single) function BackgroundCountsPerPixel2(nData, data, &
                       nSigma, sigma, sigmaPoisson)
    integer(int32), intent(in) :: nData
    real(single) :: data(:)

  
    real(single), intent(in)  :: nSigma
    real(single), intent(out) :: sigma
    real(single), intent(out) :: sigmaPoisson
    !****************************************************************************
    ! 2013-10-15 vny:
    ! modifying the calculation of the background count error: instead of sigma1 
    ! calculating sigma1/sqrt(n) where n is the number of averaged points
    ! sigma will correspond now to the sigma of the average
    !****************************************************************************

    sigma=1.0
    sigmaPoisson=1.0
    !write(*,*) "L190 backgroundMethod=",backgroundMethod 
    select case (backgroundMethod)
    case(1)   ! The current SAS
       !write(*,*) "L193 data(1)=", data(1), " ndata=", ndata
        backgroundCountsPerPixel2 = backgroundmode1(ndata, data, &
             0.001, backgroundnsigma, sigma)
        !write(*,*) "L196 backgroundCountsPerPixel2=", backgroundCountsPerPixel2, &
        !     " sigma=", sigma
        ! 2013-10-28 using Poissonian sigma as sqrt
        sigmaPoisson=sqrt(backgroundCountsPerPixel2*nData)
        !write(*,*) "L200 sigmaPoisson=", sigmaPoisson
        if (nData>1) then
           sigmaPoisson=sigmaPoisson/nData ! sigma per pixel
        endif
        !write(*,*) "L204 sigmaPoisson=", sigmaPoisson
    case(2)   ! Simple average
       !write(*,*) "L206 data(1)=", data(1), " ndata=", ndata
       call ComputeAverage2(data, ndata, backgroundCountsPerPixel2, sigma)
       !write(*,*) "L208 backgroundCountsPerPixel2=", backgroundCountsPerPixel2
       ! 2013-10-28 using Poissonian sigma as sqrt
       sigmaPoisson=sqrt(backgroundCountsPerPixel2*nData)
       if (nData>1) then
          sigmaPoisson=sigmaPoisson/nData ! sigma per pixel
       endif

    case(3)   ! Simple median
       !write(*,*) "L214 data(1)=", data(1), " ndata=", ndata
        backgroundCountsPerPixel2 = MedianValue(data, nData)
        sigma = max(1.0, sqrt(backgroundCountsPerPixel2))
        !write(*,*) "L217 backgroundCountsPerPixel2=", backgroundCountsPerPixel2
        
        ! 2013-10-28 using Poissonian sigma as sqrt
        sigmaPoisson=sqrt(backgroundCountsPerPixel2*nData)
        if (nData>1) then
           sigmaPoisson=sigmaPoisson/nData ! sigma per pixel
        endif
        
    case(4)   ! The mode
   
       !write(*,*) "L227 data(1)=", data(1), " ndata=", ndata
        backgroundCountsPerPixel2 = ommode(ndata, data)
        sigma = max(1.0, sqrt(backgroundCountsPerPixel2))
        !write(*,*) "L230 backgroundCountsPerPixel2=", backgroundCountsPerPixel2
        ! 2013-10-28 using Poissonian sigma as sqrt
        sigmaPoisson=sqrt(backgroundCountsPerPixel2*nData)
        if (nData>1) then
           sigmaPoisson=sigmaPoisson/nData ! sigma per pixel
        endif

    case(5)   ! Robust mean
       !write(*,*) "L238 data(1)=", data(1), " ndata=", ndata
       backgroundCountsPerPixel2 = RobustMean(data, nData, sigma)
       !write(*,*) "L240 backgroundCountsPerPixel2=", backgroundCountsPerPixel2
       ! 2013-10-28 using Poissonian sigma as sqrt
       sigmaPoisson=sqrt(backgroundCountsPerPixel2*nData)
       if (nData>1) then
          sigmaPoisson=sigmaPoisson/nData ! sigma per pixel
       endif

    case(6) ! Robust median
       !write(*,*) "L248 data(1)=", data(1), " ndata=", ndata
        backgroundCountsPerPixel2 = RobustMedian(data, nData)
        sigma = max(1.0, sqrt(backgroundCountsPerPixel2))
        !write(*,*) "L251 backgroundCountsPerPixel2=", backgroundCountsPerPixel2
        ! 2013-10-28 test Poissonian sigma as sqrt
        sigmaPoisson=sqrt(backgroundCountsPerPixel2*nData)
        if (nData>1) then
           sigmaPoisson=sigmaPoisson/nData ! sigma per pixel
        endif
        
    case default
        if(nsigma < 0.0) then
        end if
    end select

end function BackgroundCountsPerPixel2

!*****************************************
!
! Function BackgroundMode - Current SAS
!
!*****************************************

function BackgroundMode1(nData, data, &
                        tolerance, nSigma, sigma) result(mode)
    integer(int32), intent(in) :: nData
    real(single) :: data(:)

    real(single), intent(in) :: tolerance    ! Tolerance is usually set to 0.2
    real(single), intent(in)  :: nSigma
    real(single), intent(out) :: sigma
 

    real(single), pointer :: list(:) => null()
    integer(int32) :: nlist
   
    real(single)    :: mode
    integer(int32) :: i, k, allocErr
     
    real(single)   :: mean, median
    real(double)   :: variance, testSigma
    real(single)   :: highValue, lowValue
    integer(int32) :: m, ndataMin
    
    real(single) :: nsigma1   
!write(*, *) "mssllib backgroundmode1"



!    mode = chauvenetmode(ndata, data, sigma)





!    mode = max( 0.0_double, sum( dble(data(1 : ndata)) ) / dble(ndata))
!    sigma = max(1.0, sqrt(mode))

!    return
!****************************************************
! 
!
!***************************************************




    nList = nData
    nDataMin = nData / 2
    list => AllocateSingle(1, nList, "list (ComputeBackgroundMode)")
    list = data(1 : ndata)


    ! Sort the data so the median can be
    ! quickly computed

    call HeapSort(list, nList)


        
    do i = nList, nDataMin, -1

        median = SortedMedian(list, i)
 
        if ( (list(i) - median) < 5.0 * max(1.0, sqrt(median)) ) then
            nList = i
            exit
        end if 

    end do    

    nsigma1 = nsigma
  
     ! reject data at more than nSigma from data
   
    testSigma = 0.0

    sigmaClip: do i = 1, 20   !  was 50



      
        mean = sum( list(1 : nList)) / dble(nList)    
        median = SortedMedian(list, nlist)

        if( abs(mean - median) .le. 1.0) then
            exit
        end if
        
        if (median .le. 2.0 .and. (mean - median) .le. 1.0) then
            nsigma1 = 3.0
            median = mean
        end if

                
        !*******************************
        ! Now do sigma clipping
        ! Get variance and std deviation
        !********************************
        variance = sum(dble(list(1 : nlist) - median) ** 2)

        variance = variance / dble(nlist - 1)
        sigma = sqrt(variance) 

         
        if (sigma .le. sqrt(median)) nsigma1 = 3.0
        sigma = max(1.0, sigma)
     
        lowValue = max(0.0, median - 5.0 * sigma)
        highValue = median + nSigma1 * sigma
    
        m = ishft(nlist, -1) 
        do k = m + 1, nlist
            if(list(k) .ge. lowValue .and. list(k) .le. highValue) then

                m = m + 1            
            else
                exit   
            end if
        end do
       

        if(m .lt. ndataMin .or. sigma .lt. 0.5) exit sigmaClip
     
        if(abs((sigma - testSigma) / sigma) .lt. tolerance) exit sigmaClip
     

       
        nList = m
   
        testSigma = sigma
        
    end do sigmaClip




    mode = sum( list(1 : nList)) / dble(nList)    


    variance = sum(dble(list(1 : nlist) - mode) ** 2)
    variance = variance / dble(nlist - 1)
    sigma = sqrt(variance)



    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")




end function BackgroundMode1





!**************************************************
! ComputeBackgroundMode
! Computes the background mode (counts/per pixel)
! using the image (background + source) and 
! the source-map image (levelImage). Ignores
! any image pixels that correspond to a levelImage
! one that has a non-zero value (ie a source pixel)
! Also returns the variance of the sky-background 
!***************************************************
function ComputeBackgroundMode(image, levelImage, &
                                 nx, ny, nSigma, sigma) result(mode)

    integer(int16), intent(in)  :: nx, ny
    real(single), pointer :: image(:, :)

    integer(int16), pointer   :: levelImage(:, :)
    real(single), intent(in)    :: nSigma
    real(single), intent(out)    :: sigma
    real(single)                :: mode, median
    
    integer(int32) :: i, j
    !AI casting to a integer kind 9 to avoid multiplication
    !overflow
    integer(int32) :: a,b
    integer(int32) :: allocErr, nlist
    real(single), pointer :: list(:)
    integer :: nx1
   
    nx1 = nx - 1
    a = int(nx,int32)
    b = int(ny,int32)
    list => AllocateSingle(1, a*b, "list (Computemode)")
    
 
    nlist = 0
 
    do j = 0, ny - 1
        do i = 0, nx1

            if (levelImage(i, j) .eq. 0 .and. image(i, j) .ge. 0.0 &
                .and. .not. isNullValue(image(i, j)) ) then
                nList = nList + 1
                list(nlist) = image(i, j) 
               
            end if
        end do
    end do 


!write(*, *) "Line 361 nlist=", nlist
!write(*, *) "minval levelimage=", minval(levelImage), " max=", maxval(levelImage)

    if(nList .lt. 100) then

       
 
        nlist = 0
 
        do j = 0, ny - 1
            do i = 0, nx1

                if ( image(i, j) .ge. 0.0 .and. .not. isNullValue(image(i, j)) ) then
                    nList = nList + 1
                    list(nlist) = image(i, j) 
               
                end if
            end do
        end do 
        if(nList > 10) then
            median = MedianValue(list, nList)
            median = median + max(2.0, 2.0 * sqrt(median) ) 
            nlist = 0
 
            do j = 0, ny - 1
                do i = 0, nx1

                    if ( (image(i, j) .ge. 0.0 .and. image(i, j) .le. median) &
                         .and. .not. isNullValue(image(i, j))) then
                        nList = nList + 1
                        list(nlist) = image(i, j) 
                  
                    end if
                end do
            end do 
   

        end if
    end if

!write(*, *) "omskybackground line 348 nlist=", nlist, " minval(list)=", minval(list(1:nList)), &
!     " maxval=", maxval(list(1:nlist))   
    if (nList .eq. 0) then
        mode = 0.0
    else
        mode = BackgroundCountsPerPixel(nList, list, nSigma, sigma)  ! 


    end if
!write(*, *) "omskybackground_mod.f90 line 372 mode=", mode, " nList=", nList
 
    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
  
   
end function ComputeBackgroundMode

!**************************************************
! ComputeBackgroundMode1
! Computes the background mode counts
! using the image (background + source) 
!***************************************************
     
 
function ComputeBackgroundMode1(size, background, &
             nSigma, standardDeviation) result(mode)

    integer(int32), intent(in)  :: size
    real(single), pointer :: background(:)

    real(single), intent(in)    :: nSigma
    real(single), intent(out)   :: standardDeviation

    real(single)                :: mode
                   
    mode = BackgroundMode(size, background, 0.2, nSigma, standardDeviation) ! ws 0.1
    
end function ComputeBackgroundMode1
!**************************************************
! ComputeBackgroundMode
! Computes the background mode (counts/per pixel)
! using the image (background + source) and 
! the source-map image (levelImage). Ignores
! any image pixels that correspond to a levelImage
! one that has a non-zero value (ie a source pixel)
! Also returns the variance of the sky-background 
!***************************************************
function ComputeBackgroundMode2(image, &
                                 nx, ny, nSigma, sigma) result(mode)

    integer(int16), intent(in)  :: nx, ny
    real(single), pointer :: image(:, :)
    real(single), intent(in)    :: nSigma
    real(single), intent(out)    :: sigma
    real(single)                :: mode
    
    integer(int32) :: i, j
    integer(int32) :: a,b
    integer(int32) :: allocErr, nlist
    real(single), pointer :: list(:)
    integer :: nx1

    nx1 = nx - 1
    !AI casting to a integer kind 9 to avoid multiplication
    !overflow
    a = int(nx,int32)
    b = int(ny,int32)

    list => AllocateSingle(1, a*b, "list (Computemode)")
    
 
    nlist = 0
 
    do j = 0, ny - 1
        do i = 0, nx1

            if (image(i, j) .ge. 0.0 .and. .not. isNullValue(image(i, j)) ) then
                nList = nList + 1
                list(nlist) = image(i, j) 
            end if  
        end do
    end do 
    if(nList .gt. 0) then
        mode = BackgroundCountsPerPixel(nList, list, nSigma, sigma)  ! 
    else
        mode = 0.0
    end if
 
    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
  
   
end function ComputeBackgroundMode2

!**************************************************
! ComputeBackgroundMode
! Computes the background mode (counts/per pixel)
! using the image (background + source) and 
! the source-map image (levelImage). Ignores
! any image pixels that correspond to a levelImage
! one that has a non-zero value (ie a source pixel)
! Also returns the variance of the sky-background 
!***************************************************
function ComputeBackgroundMode3(image, sigma) result(mode)

     real(single) :: image(:, :)
 
    real(single), intent(out)    :: sigma
    real(single)                :: mode
    
    integer(int32) :: i, j

    integer(int32) :: allocErr, nlist
    real(single), pointer :: list(:)
    integer(int32) :: nx, ny
     
    nx = size(image, 1)
    ny = size(image, 2)


    list => AllocateSingle(1, nx * ny, "list (Computemode)")
  
 
    nlist = 0


!write(*, *) "Computebackgroundmode3 line 666"
!write(*, *) "lbound image, 1=", lbound(image, 1), " ubound=", ubound(image, 1)
!write(*, *) "lbound image, 2=", lbound(image, 2), " ubound=", ubound(image, 2)

 
    do j = lbound(image, 2), ubound(image, 2)
        do i = lbound(image, 1), ubound(image, 1)

            if (image(i, j) .ge. 0.0 .and. .not. isNullValue(image(i, j)) ) then           
                nList = nList + 1
                list(nlist) = image(i, j) 
            end if  
        end do
    end do 
!   write(*, *) "Line 680 nlist=", nlist, "nx=", nx, " ny=", ny
!write(*, *) "minval(list)=", minval(list), " max=", maxval(list)
 
    if(nList .gt. 0) then
        mode = BackgroundCountsPerPixel(nList, list, 2.0, sigma)  ! 
    else
        mode = 0.0
    end if

 
    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
  
   
end function ComputeBackgroundMode3


!**************************************************
! ComputeBackgroundMode
! Computes the background mode (counts/per pixel)
! using the image (background + source) and 
! the source-map image (levelImage). Ignores
! any image pixels that correspond to a levelImage
! one that has a non-zero value (ie a source pixel)
! Also returns the variance of the sky-background 
!***************************************************
function ComputeBackgroundMode4(image, levelImage, qualityImage, &
                                 nx, ny, nSigma, sigma) result(mode)

    integer(int16), intent(in)  :: nx, ny
    real(single), pointer :: image(:, :)
    
    integer(int16), pointer :: levelImage(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
    real(single), intent(in)    :: nSigma
    real(single), intent(out)    :: sigma
    real(single)                :: mode
    
    integer(int32) :: i, j, nx1
    !AI casting to allow numbers greater than 32767
    integer(int32) :: a,b
    integer(int32) :: allocErr, nlist
    real(single), pointer :: list(:)

    !AI casting to a integer kind 9 to avoid multiplication
    !overflow
    a = int(nx,int32)
    b = int(ny,int32)

    list => AllocateSingle(1,a*b, "list (Computemode)")
    
 
    nlist = 0
    nx1 = nx - 1
    do j = lbound(image, 2), ubound(image, 2)
        do i = lbound(image, 1), ubound(image, 1)

            if (levelImage(i, j) .eq. 0 .and. image(i, j) &
              .ge. 0.0 .and. .not. qualityImage(i, j)) then
                nList = nList + 1
                list(nlist) = image(i, j) 
               
            end if
        end do
    end do 

    if(nList > 0) then
        mode = BackgroundCountsPerPixel(nList, list, nSigma, sigma)  ! 
    else
        mode = 0.0
    end if

!write(*, *) "omskybackground_mod.f90 line 545 mode=", mode, " nList=", nList
 
    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")
  
   
end function ComputeBackgroundMode4


!********************************************************************
! returns an image. All pixels that belong to the same source have
! the same contents (source id)
! Modifies the level image
! If number > 0 pixels set to number
!********************************************************************
function CreateLevelImageMap(nx, ny, sources) result(levelImage)
    integer(int16), intent(in) :: nx, ny
    type(SourceParameters), pointer :: sources(:)
    integer(int16),          pointer    :: levelImage(:, :)   
    real(single)   :: sourceRadius, radius, x0, y0, xPixel, yPixel
    integer(int32)        :: r, i, j, k, i0, j0, i1, i2, j1, j2
    
    
    sourceRadius = real(10 / binX)


    r = ceiling(sourceRadius)    
    levelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "levelImage")
    if (.not. associated(sources)) return

    do k = 1, size(sources)
        if (sources(k)%extensionFlag .eq. 1) cycle
       

  
        x0 = sources(k)%xPos
        y0 = sources(k)%yPos
        i0 = int(x0)
        j0 = int(y0)


        ! Determine upper and lower array indices for summing counts
        i1 = i0 - r - 1
        i1 = max(0, i1)

        i2 = i0 + r + 1
        i2 = min(nx - 1, i2)

        j1 = j0 - r - 1
        j1 = max(0, j1)

        j2 = j0 + r + 1
        j2 = min(ny - 1, j2)

        do i = i1, i2
            xPixel = int(i)
            do j = j1, j2               
                yPixel = int(j)
                radius = sqrt((xPixel - x0) ** 2 + (yPixel - y0) ** 2)

                if (radius .le. sourceRadius) then
                    levelImage(i, j) = k
                end if

            end do
        end do

    end do
   
end function CreateLevelImageMap




!*****************************************
! Function experimentalsasMode
!*****************************************

function experimentalSasMode(nData, data, sigma) result(mean)
    integer(int32), intent(in) :: nData
    real(single) :: data(:)

 
    real(single), intent(out) :: sigma
 

    real(single), pointer :: list(:) => null()
    integer(int32) :: nlist
  
    integer(int32) :: i
     
    real(single)   :: mean, median
   
    integer(int32) :: ndataMin
    real(single) :: nsigma1    
   
  
  
    mean = sum(data(1 : ndata)) / dble(ndata)



    if (mean <= 1.0) return





    
    nList = nData
    nDataMin = max(20, nData / 4)
    list => AllocateSingle(1, nList, "list (ComputeBackgroundMode)")
    list = data(1 : ndata)


    ! Sort the data so the median can be
    ! quickly computed

    call HeapSort(list, nList)


    median = SortedMedian(list, nlist)
    if (abs(mean - median) / mean < 0.01) return

       



    nsigma1 = 3.0



    do i = nList, nDataMin, -1






        mean = sum( list(1 : i)) / dble(i)    
       
        if (mean < 0.5) return
        


        median = SortedMedian(list, i)

        if (abs(mean - median) / mean < 0.2 .or. mean < 3.0) nsigma1 = 4.0

        median = min(mean, median)
        
                 
    
        if ( (list(i) - median) < nsigma1 * max(1.0, sqrt(median)) ) then
            return
        end if 
        

    end do    


    sigma = 0.0

end function experimentalSasMode








!*****************************************
! Function ChauvenetMode
!*****************************************

function chauvenetMode(nData, data, sigma) result(mean)
    integer(int32), intent(in) :: nData
    real(single) :: data(:)

    real(single), intent(out) :: sigma

    real(single), pointer :: list(:) => null()
    integer(int32) :: nlist
    integer(int32) :: i, ilOld, il, iu, iuOld, j, k, n

    real(single)   :: mean, median, lowValue, highValue, r
    real(double)   :: variance, sum1, sum2, diff
    integer(int32) :: ndataMin
!    integer, save :: count = 0
 
    mean = sum( data(1 : nData) ) / dble(nData)


 

    if (mean < 1.0) return

    nList = nData
    nDataMin = max(20, nData / 4)
    list => AllocateSingle(1, nList, "list (ComputeBackgroundMode)")
    list = data(1 : ndata)


    ! Sort the data so the median can be
    ! quickly computed

    call HeapSort(list, nList)

!    do i = 1, nlist
!      write(*, *) i, " pixel=", list(i)
!    end do




    nList = 0    
    do i = 1, nData

            
        if ( list(i) >=  0.0 ) then
            nList = nlist + 1
            list(nlist) = list(i)
        end if 

    end do    


    if (nlist < 1) then
        mean = 0.0
        deallocate(list) 
        return
    end if


    median = max(1.0,SortedMedian(list, nlist))
    highValue = max(6.0, median + 6.0 * max(1.0, sqrt(median)))
    k = 0    
    do i = 1, nList

       
        if ( list(i) > highValue ) exit
        k = k + 1
        list(k) = list(i)


    end do    

    if (k < 1) then
        mean = 0.0
        deallocate(list) 
        return
    end if
    nlist = k


    ! reject data at more than nSigma from data
!write(*, *) "*************************************************************"
!write(*, *) "Chauvenet mode initial median =", median, " ndata=", ndata, " nlist=", nlist
   
    il = 1
    iu = nList
    ilOld = il
    iuOld = iu

    do j = 1, 20
       
      
        n = iu - il + 1
        mean = sum( list(il : iu)) / dble(n)
        if (mean < 1.0) then


            exit
        end if 
        
        median = SortedMedian(list(il : iu), n)
         
        sum1 = 0.0_double
        sum2 = 0.0_double
        
        do i = il, iu
            diff = dble(data(i)) - dble(median)
            sum1 = sum1 + diff
            sum2 = sum2 + diff * diff
        end do
                     
        variance = ( sum2 - sum1 * sum1  / dble(n) ) / dble( n - 1)
        
        sigma = max(1.0_double, sqrt(variance))
!        sigma = max(1.0, min( sigma, 2.0 * sqrt(median)))        

        !************************************************************
        ! Compute Chauvenet rejection criterion
        !***********************************************************
        r = log10( real(n) )
        r = max( 3.0, (-0.1042 * r + 1.1695) * r + 0.8895) ! was 3.0, 
        lowValue = median - r * sigma

        highValue = max(median + 5.0, median + r * sigma)

!        write(*, *) "Iteration ", j, " il=", il, " iu=", iu, " mean=", mean, &
!           " median=", median, " mode=", mode, " sigma=", sigma, " r=", r, &
!           " lowvalue=", lowvalue, " high=", highvalue, " n=", n 
!        write(*, *) "std=", sqrt(sum2 / dble(n - 1))

        do i = 1, nData   
            if (list(i) >= lowValue) then
                il = i
                exit
           
            end if 
        end do

        do i = nData, 1, -1   
            if (list(i) <= highValue) then
                iu = i         
                exit
            end if 
        end do
    
        if (mean < median) exit
 
        
        if (il .eq. ilOld .and. iu .eq. iuOld) exit
        ilOld = il
        iuOld = iu

    end do    
!write(*, *) "Final list mean=", mean, " il=", il, " iu=", iu
!count = count + 1

!if (sigma > 20.0) then
!write(*, *) "Final pixel list sigma=", sigma
!  do i = il, iu
!      write(*, *) i, " pixel=", list(i)
!    end do
!end if
!if (count == 20) call fatal("stopped")

    deallocate(list)
   
end function chauvenetmode





!*****************************************
! Function BackgroundMode
!*****************************************

function BackgroundMode100(nData, data, sigma) result(mode)
    integer(int32), intent(in) :: nData
    real(single) :: data(:)

  
    real(single), intent(out) :: sigma
 

    real(single), pointer :: list(:) => null()
    integer(int32) :: nlist
   
    real(single)    :: mode
    integer(int32) :: i, j, k, allocErr
     
    real(single)   :: mean, median
   
    integer(int32) :: ndataMin
    


    real(double) :: ax, bx, tol, xmin, fmin


    nList = nData

    nDataMin = nData / 2
    list => AllocateSingle(1, nList, "list (ComputeBackgroundMode)")
    list = data(1 : ndata)

   



    ! Sort the data so the median can be
    ! quickly computed
    call HeapSort(list, nList)



    median = SortedMedian(list, nlist)
 
    if (median < 1.0) then
    

        mode = sum( dble(list(1 : nlist) )) / dble(nlist) 

        sigma = sum(dble(list(1 : nlist) - mode) ** 2)
        sigma = sigma / dble(nlist - 1)
        sigma = sqrt(sigma)


    
        return


    end if





 
  




    
  
     ! reject data at more than nSigma from data
   
   

    do i = 1, 20   !  was 50
        
      
        mean = sum( list(1 : nList)) / dble(nList) 
        if (mean < 0.5) then  
    
         
            exit
        end if



        median = SortedMedian(list, nlist)
       
        if ( (mean - median) / mean < 0.25) then

        
            median = mean
        end if

        if (abs(mean - median) <= 1.0) median = mean

                
       
  
        meanValue = dble(median)       
        ax = max(1.0, median)
        bx = ax + 10.0 * max(1.0_double, sqrt(ax) )
        tol = 0.00001_double 
        fmin = BRENT( ax, bx, tol, xmin, probFunction)
        k = nlist

        do j = nlist, ndataMin, -1
            if(data(j) .le. xmin) then
                k = j
                exit
            end if 
        end do

        if (k .eq. nList) then 
        
            exit      
        end if

        if(k .le. ndataMin) then

         
            exit      
        end if
 
          
        nList = k
   
     
        
    end do 
  

    mode = sum( list(1 : nList)) / dble(nList)    


    sigma = sum(dble(list(1 : nlist) - mode) ** 2)
    sigma = sigma / dble(nlist - 1)
    sigma = sqrt(sigma)




    deallocate(list, stat = allocErr)
    call CheckDeallocation(allocErr, "list")




end function BackgroundMode100




!******************************************************
!
!*****************************************************
real(double) function probFunction( x )
    real(double), intent(in) :: x
    real(double), parameter :: prob = 0.9925

    probFunction = dabs( gammap(meanValue, x) - prob )

end function probFunction





end module Omskybackground_mod









